//* File: main-psp.c */

/*
 * Copyright (c) 1997 Ben Harrison
 *
 * This software may be copied and distributed for educational, research,
 * and not for profit purposes provided that this copyright and statement
 * are included in all such copies.
 */


/*
 * This file helps Angband work on the Sony Playstation Portable
 *
 * To use this file, use "Makefile.psp", which defines USE_PSP.
 */


#include "angband.h"

#ifdef USE_PSP

#include <pspkernel.h>
#include <pspctrl.h>
#include <pspdebug.h>
#include <pspdisplay.h>
#include <pspge.h>

#define PSP_SCREEN_WIDTH 480
#define PSP_SCREEN_HEIGHT 272
#define PSP_LINE_SIZE 512
#define PSP_PIXEL_FORMAT 3

#include "main.h"
#define INTERNAL_MAIN

PSP_MODULE_INFO("Angband", 0, 1, 1);
PSP_MAIN_THREAD_ATTR(0);

typedef struct star_t_struct {
	struct star_t_struct* links[4];
	char macro[64];
	char desc[64];
} star_t;

// 3 display routines for the psp (slightly modified versions of the pspsdk ones)
void psp_screen_init();
void psp_put_char(int x, int y, unsigned long color, unsigned long bk_color, unsigned char ch);
void psp_put_string(int x, int y, unsigned long color, unsigned long bk_color, char* string);

// Thread management routines
int psp_exit_callback(void); // exit the program on the home button
void psp_callback_thread(void *arg); // thread to run the callback functions
void psp_timer_thread(void* arg); // thread to blink the cursor
void psp_input_thread(void* arg); // thread to manage user input

void psp_cursor_flip(); // draws the cursor

// General input routines
int psp_analog_octant(int x, int y); // maps the analog stick to an 8 way pad
void psp_update_keys(); // gets the current button status

// Onscreen keyboard routines
void psp_show_osk(); // displays the onscreen keyboard
void psp_highlight_osk(int cx, int cy, int color); // highlight one part of the keyboard
void psp_hide_osk(); // erases the onscreen keyboard
char psp_do_osk(); // get a character from the user via the onscreen keyboard

// Macro routines
void psp_show_star(star_t* star); // displays the macro interface 
void psp_hide_star(); // erases the macro interface
void psp_save_star(int fff, star_t* star); // recursively saves a macro star to disk
void psp_save_macros(); // saves all macros to disk
void psp_load_star(int fff, star_t* star); // recursively loads a star from disk
void psp_load_macros(); // loads all macros from disk
void psp_modify_macro(); // allows the user to change macros and save the changes in game
void psp_do_macro(); // invokes a macro selected by the user

// The standard Angband terminal hooks
static void Term_init_psp(term *t);
static void Term_nuke_psp(term *t);
static errr Term_user_psp(int n);
static errr Term_xtra_psp(int n, int v);
static errr Term_curs_psp(int x, int y);
static errr Term_wipe_psp(int x, int y, int n);
static errr Term_text_psp(int x, int y, int n, byte a, const char *cp);
static errr Term_pict_psp(int x, int y, int n, const byte *ap, const char *cp,
                          const byte *tap, const char *tcp);
static void term_data_link(int i);

int main_thid, input_thid, timer_thid;

// The palette to display text in 
static long pal[16] = {
				0x00000000, /* Black */
				0x00FFFFFF, /* White */
				0x00808080, /* Gray */
				0x000080FF, /* Orange */
				0x000000FF, /* Red */
				0x00408000, /* Green */
				0x00FF0000, /* Blue */
				0x00004080, /* Brown */
				0x00606060, /* Dark Gray */
				0x00A0A0A0, /* Light Gray */
				0x00FF00FF, /* Purple */
				0x0000FFFF, /* Yellow */
				0x004040FF, /* Light Red */
				0x0000FF00, /* Light Green */
				0x00FFFF00, /* Light Blue */
				0x004080C0  /* Light Brown */
};

unsigned long* vram_base;
bool screen_init = 0;

#define NUM_KEYS 12
int key_signals[NUM_KEYS] = {PSP_CTRL_SELECT, PSP_CTRL_START, PSP_CTRL_UP, PSP_CTRL_RIGHT,
				PSP_CTRL_DOWN, PSP_CTRL_LEFT, PSP_CTRL_LTRIGGER, PSP_CTRL_RTRIGGER,
				PSP_CTRL_TRIANGLE, PSP_CTRL_CIRCLE, PSP_CTRL_CROSS, PSP_CTRL_SQUARE};
enum {KEY_SELECT, KEY_START, KEY_UP, KEY_RIGHT, KEY_DOWN, KEY_LEFT, 
			KEY_LTRIGGER, KEY_RTRIGGER, KEY_TRIANGLE, KEY_CIRCLE, KEY_CROSS, KEY_SQUARE};
enum {STICK_C, STICK_N, STICK_NE, STICK_E, STICK_SE, 
				STICK_S, STICK_SW, STICK_W, STICK_NW};

#define HELD(x) (key_state[x] != -1)
#define PRESSED(x) (!key_state[x] || (key_state[x] > 20 && !(key_state[x] % 4)))
#define A_HELD(x) (analog_octant == x)
#define A_PRESSED(x) (analog_octant == x && (!analog_state || (!(analog_state % 6) && analog_state >= 12)))

char keys[2][52] = {"`1234567890-=abcdefghijklmnopqrstuvwxyz[]\\;',./    ",
									"~!@#$%^&*()_+ABCDEFGHIJKLMNOPQRSTUVWXYZ{}|\":<>?     "};
int key_state[NUM_KEYS];
int analog_state = 0;
char analog_octant = STICK_C;

// The default macros
star_t star_ll = {{NULL, NULL, NULL, NULL}, "mbf", "Teleport Self"};
star_t star_lu = {{NULL, NULL, NULL, NULL}, "mab", "Detect Monsters"};
star_t star_lr = {{NULL, NULL, NULL, NULL}, "mcf", "Identify"};
star_t star_ld = {{NULL, NULL, NULL, NULL}, "mah", "Detect Traps/Doors"};
star_t star_ul = {{NULL, NULL, NULL, NULL}, "R&\r", "Rest As Needed"};
star_t star_uu = {{NULL, NULL, NULL, NULL}, "R100\r", "Rest 100 Turns"};
star_t star_ur = {{NULL, NULL, NULL, NULL}, "maf", "Cure Light Wounds"};
star_t star_ud = {{NULL, NULL, NULL, NULL}, "mad", "Light Area"};
star_t star_rl = {{NULL, NULL, NULL, NULL}, "mcd", "Fire Bolt"};
star_t star_ru = {{NULL, NULL, NULL, NULL}, "mai", "Stinking Cloud"};
star_t star_rr = {{NULL, NULL, NULL, NULL}, "mbb", "Lightning Bolt"};
star_t star_rd = {{NULL, NULL, NULL, NULL}, "mda", "Frost Ball"};
star_t star_dl = {{NULL, NULL, NULL, NULL}, "k-y", "Destroy Ground Object"};
star_t star_du = {{NULL, NULL, NULL, NULL}, "0100s", "Search"};
star_t star_dr = {{NULL, NULL, NULL, NULL}, "mcc", "Stone To Mud"};
star_t star_dd = {{NULL, NULL, NULL, NULL}, "0100D", "Disarm"};
star_t star_l  = {{&star_ll, &star_lu, &star_lr, &star_ld}, "mac", "Phase Door"};
star_t star_u  = {{&star_ul, &star_uu, &star_ur, &star_ud}, "R*\r", "Rest HP/MP"};
star_t star_r  = {{&star_rl, &star_ru, &star_rr, &star_rd}, "maa", "M. Missile"};
star_t star_d  = {{&star_dl, &star_du, &star_dr, &star_dd}, "0100o", "Pick Lock"};
star_t star    = {{&star_l, &star_u, &star_r, &star_d}, "*tmaa5", "M. Missile Nearest"};

// A font that is 99% the same as the one from the ps2dev.org pspsdk.  It
//   has the psp button symbols (0x07 - 0x0A) and a 50% hash block (0xB1) though
u8 font[]=
"\x00\x00\x00\x00\x00\x00\x00\x00\x3c\x42\xa5\x81\xa5\x99\x42\x3c"
"\x3c\x7e\xdb\xff\xff\xdb\x66\x3c\x6c\xfe\xfe\xfe\x7c\x38\x10\x00"
"\x10\x38\x7c\xfe\x7c\x38\x10\x00\x10\x38\x54\xfe\x54\x10\x38\x00"
"\x10\x38\x7c\xfe\xfe\x10\x38\x00\x00\x70\x88\x88\x88\x70\x00\x00"
"\x00\x20\x55\x55\x88\xf8\x00\x00\x00\xf8\x88\x88\x88\xf8\x00\x00"
"\x00\x88\x50\x20\x50\x88\x00\x00\x0f\x03\x05\x79\x88\x88\x88\x70"
"\x38\x44\x44\x44\x38\x10\x7c\x10\x30\x28\x24\x24\x28\x20\xe0\xc0"
"\x3c\x24\x3c\x24\x24\xe4\xdc\x18\x10\x54\x38\xee\x38\x54\x10\x00"
"\x10\x10\x10\x7c\x10\x10\x10\x10\x10\x10\x10\xff\x00\x00\x00\x00"
"\x00\x00\x00\xff\x10\x10\x10\x10\x10\x10\x10\xf0\x10\x10\x10\x10"
"\x10\x10\x10\x1f\x10\x10\x10\x10\x10\x10\x10\xff\x10\x10\x10\x10"
"\x10\x10\x10\x10\x10\x10\x10\x10\x00\x00\x00\xff\x00\x00\x00\x00"
"\x00\x00\x00\x1f\x10\x10\x10\x10\x00\x00\x00\xf0\x10\x10\x10\x10"
"\x10\x10\x10\x1f\x00\x00\x00\x00\x10\x10\x10\xf0\x00\x00\x00\x00"
"\x81\x42\x24\x18\x18\x24\x42\x81\x01\x02\x04\x08\x10\x20\x40\x80"
"\x80\x40\x20\x10\x08\x04\x02\x01\x00\x10\x10\xff\x10\x10\x00\x00"
"\x00\x00\x00\x00\x00\x00\x00\x00\x20\x20\x20\x20\x00\x00\x20\x00"
"\x50\x50\x50\x00\x00\x00\x00\x00\x50\x50\xf8\x50\xf8\x50\x50\x00"
"\x20\x78\xa0\x70\x28\xf0\x20\x00\xc0\xc8\x10\x20\x40\x98\x18\x00"
"\x40\xa0\x40\xa8\x90\x98\x60\x00\x10\x20\x40\x00\x00\x00\x00\x00"
"\x10\x20\x40\x40\x40\x20\x10\x00\x40\x20\x10\x10\x10\x20\x40\x00"
"\x20\xa8\x70\x20\x70\xa8\x20\x00\x00\x20\x20\xf8\x20\x20\x00\x00"
"\x00\x00\x00\x00\x00\x20\x20\x40\x00\x00\x00\x78\x00\x00\x00\x00"
"\x00\x00\x00\x00\x00\x60\x60\x00\x00\x00\x08\x10\x20\x40\x80\x00"
"\x70\x88\x98\xa8\xc8\x88\x70\x00\x20\x60\xa0\x20\x20\x20\xf8\x00"
"\x70\x88\x08\x10\x60\x80\xf8\x00\x70\x88\x08\x30\x08\x88\x70\x00"
"\x10\x30\x50\x90\xf8\x10\x10\x00\xf8\x80\xe0\x10\x08\x10\xe0\x00"
"\x30\x40\x80\xf0\x88\x88\x70\x00\xf8\x88\x10\x20\x20\x20\x20\x00"
"\x70\x88\x88\x70\x88\x88\x70\x00\x70\x88\x88\x78\x08\x10\x60\x00"
"\x00\x00\x20\x00\x00\x20\x00\x00\x00\x00\x20\x00\x00\x20\x20\x40"
"\x18\x30\x60\xc0\x60\x30\x18\x00\x00\x00\xf8\x00\xf8\x00\x00\x00"
"\xc0\x60\x30\x18\x30\x60\xc0\x00\x70\x88\x08\x10\x20\x00\x20\x00"
"\x70\x88\x08\x68\xa8\xa8\x70\x00\x20\x50\x88\x88\xf8\x88\x88\x00"
"\xf0\x48\x48\x70\x48\x48\xf0\x00\x30\x48\x80\x80\x80\x48\x30\x00"
"\xe0\x50\x48\x48\x48\x50\xe0\x00\xf8\x80\x80\xf0\x80\x80\xf8\x00"
"\xf8\x80\x80\xf0\x80\x80\x80\x00\x70\x88\x80\xb8\x88\x88\x70\x00"
"\x88\x88\x88\xf8\x88\x88\x88\x00\x70\x20\x20\x20\x20\x20\x70\x00"
"\x38\x10\x10\x10\x90\x90\x60\x00\x88\x90\xa0\xc0\xa0\x90\x88\x00"
"\x80\x80\x80\x80\x80\x80\xf8\x00\x88\xd8\xa8\xa8\x88\x88\x88\x00"
"\x88\xc8\xc8\xa8\x98\x98\x88\x00\x70\x88\x88\x88\x88\x88\x70\x00"
"\xf0\x88\x88\xf0\x80\x80\x80\x00\x70\x88\x88\x88\xa8\x90\x68\x00"
"\xf0\x88\x88\xf0\xa0\x90\x88\x00\x70\x88\x80\x70\x08\x88\x70\x00"
"\xf8\x20\x20\x20\x20\x20\x20\x00\x88\x88\x88\x88\x88\x88\x70\x00"
"\x88\x88\x88\x88\x50\x50\x20\x00\x88\x88\x88\xa8\xa8\xd8\x88\x00"
"\x88\x88\x50\x20\x50\x88\x88\x00\x88\x88\x88\x70\x20\x20\x20\x00"
"\xf8\x08\x10\x20\x40\x80\xf8\x00\x70\x40\x40\x40\x40\x40\x70\x00"
"\x00\x00\x80\x40\x20\x10\x08\x00\x70\x10\x10\x10\x10\x10\x70\x00"
"\x20\x50\x88\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf8\x00"
"\x40\x20\x10\x00\x00\x00\x00\x00\x00\x00\x70\x08\x78\x88\x78\x00"
"\x80\x80\xb0\xc8\x88\xc8\xb0\x00\x00\x00\x70\x88\x80\x88\x70\x00"
"\x08\x08\x68\x98\x88\x98\x68\x00\x00\x00\x70\x88\xf8\x80\x70\x00"
"\x10\x28\x20\xf8\x20\x20\x20\x00\x00\x00\x68\x98\x98\x68\x08\x70"
"\x80\x80\xf0\x88\x88\x88\x88\x00\x20\x00\x60\x20\x20\x20\x70\x00"
"\x10\x00\x30\x10\x10\x10\x90\x60\x40\x40\x48\x50\x60\x50\x48\x00"
"\x60\x20\x20\x20\x20\x20\x70\x00\x00\x00\xd0\xa8\xa8\xa8\xa8\x00"
"\x00\x00\xb0\xc8\x88\x88\x88\x00\x00\x00\x70\x88\x88\x88\x70\x00"
"\x00\x00\xb0\xc8\xc8\xb0\x80\x80\x00\x00\x68\x98\x98\x68\x08\x08"
"\x00\x00\xb0\xc8\x80\x80\x80\x00\x00\x00\x78\x80\xf0\x08\xf0\x00"
"\x40\x40\xf0\x40\x40\x48\x30\x00\x00\x00\x90\x90\x90\x90\x68\x00"
"\x00\x00\x88\x88\x88\x50\x20\x00\x00\x00\x88\xa8\xa8\xa8\x50\x00"
"\x00\x00\x88\x50\x20\x50\x88\x00\x00\x00\x88\x88\x98\x68\x08\x70"
"\x00\x00\xf8\x10\x20\x40\xf8\x00\x18\x20\x20\x40\x20\x20\x18\x00"
"\x20\x20\x20\x00\x20\x20\x20\x00\xc0\x20\x20\x10\x20\x20\xc0\x00"
"\x40\xa8\x10\x00\x00\x00\x00\x00\x00\x00\x20\x50\xf8\x00\x00\x00"
"\x70\x88\x80\x80\x88\x70\x20\x60\x90\x00\x00\x90\x90\x90\x68\x00"
"\x10\x20\x70\x88\xf8\x80\x70\x00\x20\x50\x70\x08\x78\x88\x78\x00"
"\x48\x00\x70\x08\x78\x88\x78\x00\x20\x10\x70\x08\x78\x88\x78\x00"
"\x20\x00\x70\x08\x78\x88\x78\x00\x00\x70\x80\x80\x80\x70\x10\x60"
"\x20\x50\x70\x88\xf8\x80\x70\x00\x50\x00\x70\x88\xf8\x80\x70\x00"
"\x20\x10\x70\x88\xf8\x80\x70\x00\x50\x00\x00\x60\x20\x20\x70\x00"
"\x20\x50\x00\x60\x20\x20\x70\x00\x40\x20\x00\x60\x20\x20\x70\x00"
"\x50\x00\x20\x50\x88\xf8\x88\x00\x20\x00\x20\x50\x88\xf8\x88\x00"
"\x10\x20\xf8\x80\xf0\x80\xf8\x00\x00\x00\x6c\x12\x7e\x90\x6e\x00"
"\x3e\x50\x90\x9c\xf0\x90\x9e\x00\x60\x90\x00\x60\x90\x90\x60\x00"
"\x90\x00\x00\x60\x90\x90\x60\x00\x40\x20\x00\x60\x90\x90\x60\x00"
"\x40\xa0\x00\xa0\xa0\xa0\x50\x00\x40\x20\x00\xa0\xa0\xa0\x50\x00"
"\x90\x00\x90\x90\xb0\x50\x10\xe0\x50\x00\x70\x88\x88\x88\x70\x00"
"\x50\x00\x88\x88\x88\x88\x70\x00\x20\x20\x78\x80\x80\x78\x20\x20"
"\x18\x24\x20\xf8\x20\xe2\x5c\x00\x88\x50\x20\xf8\x20\xf8\x20\x00"
"\xc0\xa0\xa0\xc8\x9c\x88\x88\x8c\x18\x20\x20\xf8\x20\x20\x20\x40"
"\x10\x20\x70\x08\x78\x88\x78\x00\x10\x20\x00\x60\x20\x20\x70\x00"
"\x20\x40\x00\x60\x90\x90\x60\x00\x20\x40\x00\x90\x90\x90\x68\x00"
"\x50\xa0\x00\xa0\xd0\x90\x90\x00\x28\x50\x00\xc8\xa8\x98\x88\x00"
"\x00\x70\x08\x78\x88\x78\x00\xf8\x00\x60\x90\x90\x90\x60\x00\xf0"
"\x20\x00\x20\x40\x80\x88\x70\x00\x00\x00\x00\xf8\x80\x80\x00\x00"
"\x00\x00\x00\xf8\x08\x08\x00\x00\x84\x88\x90\xa8\x54\x84\x08\x1c"
"\x84\x88\x90\xa8\x58\xa8\x3c\x08\x20\x00\x00\x20\x20\x20\x20\x00"
"\x00\x00\x24\x48\x90\x48\x24\x00\x00\x00\x90\x48\x24\x48\x90\x00"
"\x28\x50\x20\x50\x88\xf8\x88\x00\x55\xAA\x55\xAA\x55\xAA\x55\xAA"
"\x28\x50\x00\x70\x20\x20\x70\x00\x28\x50\x00\x20\x20\x20\x70\x00"
"\x28\x50\x00\x70\x88\x88\x70\x00\x50\xa0\x00\x60\x90\x90\x60\x00"
"\x28\x50\x00\x88\x88\x88\x70\x00\x50\xa0\x00\xa0\xa0\xa0\x50\x00"
"\xfc\x48\x48\x48\xe8\x08\x50\x20\x00\x50\x00\x50\x50\x50\x10\x20"
"\xc0\x44\xc8\x54\xec\x54\x9e\x04\x10\xa8\x40\x00\x00\x00\x00\x00"
"\x00\x20\x50\x88\x50\x20\x00\x00\x88\x10\x20\x40\x80\x28\x00\x00"
"\x7c\xa8\xa8\x68\x28\x28\x28\x00\x38\x40\x30\x48\x48\x30\x08\x70"
"\x00\x00\x00\x00\x00\x00\xff\xff\xf0\xf0\xf0\xf0\x0f\x0f\x0f\x0f"
"\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00"
"\x00\x00\x00\x3c\x3c\x00\x00\x00\xff\xff\xff\xff\xff\xff\x00\x00"
"\xc0\xc0\xc0\xc0\xc0\xc0\xc0\xc0\x0f\x0f\x0f\x0f\xf0\xf0\xf0\xf0"
"\xfc\xfc\xfc\xfc\xfc\xfc\xfc\xfc\x03\x03\x03\x03\x03\x03\x03\x03"
"\x3f\x3f\x3f\x3f\x3f\x3f\x3f\x3f\x11\x22\x44\x88\x11\x22\x44\x88"
"\x88\x44\x22\x11\x88\x44\x22\x11\xfe\x7c\x38\x10\x00\x00\x00\x00"
"\x00\x00\x00\x00\x10\x38\x7c\xfe\x80\xc0\xe0\xf0\xe0\xc0\x80\x00"
"\x01\x03\x07\x0f\x07\x03\x01\x00\xff\x7e\x3c\x18\x18\x3c\x7e\xff"
"\x81\xc3\xe7\xff\xff\xe7\xc3\x81\xf0\xf0\xf0\xf0\x00\x00\x00\x00"
"\x00\x00\x00\x00\x0f\x0f\x0f\x0f\x0f\x0f\x0f\x0f\x00\x00\x00\x00"
"\x00\x00\x00\x00\xf0\xf0\xf0\xf0\x33\x33\xcc\xcc\x33\x33\xcc\xcc"
"\x00\x20\x20\x50\x50\x88\xf8\x00\x20\x20\x70\x20\x70\x20\x20\x00"
"\x00\x00\x00\x50\x88\xa8\x50\x00\xff\xff\xff\xff\xff\xff\xff\xff"
"\x00\x00\x00\x00\xff\xff\xff\xff\xf0\xf0\xf0\xf0\xf0\xf0\xf0\xf0"
"\x0f\x0f\x0f\x0f\x0f\x0f\x0f\x0f\xff\xff\xff\xff\x00\x00\x00\x00"
"\x00\x00\x68\x90\x90\x90\x68\x00\x30\x48\x48\x70\x48\x48\x70\xc0"
"\xf8\x88\x80\x80\x80\x80\x80\x00\xf8\x50\x50\x50\x50\x50\x98\x00"
"\xf8\x88\x40\x20\x40\x88\xf8\x00\x00\x00\x78\x90\x90\x90\x60\x00"
"\x00\x50\x50\x50\x50\x68\x80\x80\x00\x50\xa0\x20\x20\x20\x20\x00"
"\xf8\x20\x70\xa8\xa8\x70\x20\xf8\x20\x50\x88\xf8\x88\x50\x20\x00"
"\x70\x88\x88\x88\x50\x50\xd8\x00\x30\x40\x40\x20\x50\x50\x50\x20"
"\x00\x00\x00\x50\xa8\xa8\x50\x00\x08\x70\xa8\xa8\xa8\x70\x80\x00"
"\x38\x40\x80\xf8\x80\x40\x38\x00\x70\x88\x88\x88\x88\x88\x88\x00"
"\x00\xf8\x00\xf8\x00\xf8\x00\x00\x20\x20\xf8\x20\x20\x00\xf8\x00"
"\xc0\x30\x08\x30\xc0\x00\xf8\x00\x18\x60\x80\x60\x18\x00\xf8\x00"
"\x10\x28\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\xa0\x40"
"\x00\x20\x00\xf8\x00\x20\x00\x00\x00\x50\xa0\x00\x50\xa0\x00\x00"
"\x00\x18\x24\x24\x18\x00\x00\x00\x00\x30\x78\x78\x30\x00\x00\x00"
"\x00\x00\x00\x00\x30\x00\x00\x00\x3e\x20\x20\x20\xa0\x60\x20\x00"
"\xa0\x50\x50\x50\x00\x00\x00\x00\x40\xa0\x20\x40\xe0\x00\x00\x00"
"\x00\x38\x38\x38\x38\x38\x38\x00\x00\x00\x00\x00\x00\x00\x00";


/*
 * Extra data to associate with each "window"
 *
 * Each "window" is represented by a "term_data" structure, which
 * contains a "term" structure, which contains a pointer (t->data)
 * back to the term_data structure.
 */

typedef struct term_data term_data;

struct term_data
{
	term t;

	/* Other fields if needed XXX XXX XXX */
};



/*
 * Number of "term_data" structures to support XXX XXX XXX
 *
 * You MUST support at least one "term_data" structure, and the
 * game will currently use up to eight "term_data" structures if
 * they are available.
 *
 * Actually, MAX_TERM_DATA is now defined as eight in 'defines.h'.
 *
 * If only one "term_data" structure is supported, then a lot of
 * the things that would normally go into a "term_data" structure
 * could be made into global variables instead.
 */
#define MAX_PSP_TERM 1


/*
 * An array of "term_data" structures, one for each "sub-window"
 */
static term_data data[MAX_PSP_TERM];


#if 0

/*
 * Often, it is helpful to create an array of "color data" containing
 * a representation of the "angband_color_table" array in some "local" form.
 *
 * Often, the "Term_xtra(TERM_XTRA_REACT, 0)" hook is used to initialize
 * "color_data" from "angband_color_table".  XXX XXX XXX
 */
static local_color_data_type color_data[256];

#endif



/*** Function hooks needed by "Term" ***/


/*
 * Init a new "term"
 *
 * This function should do whatever is necessary to prepare a new "term"
 * for use by the "z-term.c" package.  This may include clearing the window,
 * preparing the cursor, setting the font/colors, etc.  Usually, this
 * function does nothing, and the "init_xxx()" function does it all.
 */
static void Term_init_psp(term *t)
{
	//term_data *td = (term_data*)(t->data);
	/* XXX XXX XXX */
}



/*
 * Nuke an old "term"
 *
 * This function is called when an old "term" is no longer needed.  It should
 * do whatever is needed to clean up before the program exits, such as wiping
 * the screen, restoring the cursor, fixing the font, etc.  Often this function
 * does nothing and lets the operating system clean up when the program quits.
 */
static void Term_nuke_psp(term *t)
{
	//term_data *td = (term_data*)(t->data);
	/* XXX XXX XXX */
}



/*
 * Do a "user action" on the current "term"
 *
 * This function allows the visual module to do implementation defined
 * things when the user activates the "system defined command" command.
 *
 * This function is normally not used.
 *
 * In general, this function should return zero if the action is successfully
 * handled, and non-zero if the action is unknown or incorrectly handled.
 */
static errr Term_user_psp(int n)
{
//	term_data *td = (term_data*)(Term->data);

	/* XXX XXX XXX */
	/* Unknown */
	return (1);
}


/*
 * Do a "special thing" to the current "term"
 *
 * This function must react to a large number of possible arguments, each
 * corresponding to a different "action request" by the "z-term.c" package,
 * or by the application itself.
 *
 * The "action type" is specified by the first argument, which must be a
 * constant of the form "TERM_XTRA_*" as given in "z-term.h", and the second
 * argument specifies the "information" for that argument, if any, and will
 * vary according to the first argument.
 *
 * In general, this function should return zero if the action is successfully
 * handled, and non-zero if the action is unknown or incorrectly handled.
 */
static errr Term_xtra_psp(int n, int v)
{
	//term_data *td = (term_data*)(Term->data);
	/* Analyze */
	switch (n)
	{
		case TERM_XTRA_EVENT:
		{
			/*
			 * Process some pending events XXX XXX XXX
			 *
			 * Wait for at least one event if "v" is non-zero
			 * otherwise, if no events are ready, return at once.
			 * When "keypress" events are encountered, the "ascii"
			 * value corresponding to the key should be sent to the
			 * "Term_keypress()" function.  Certain "bizarre" keys,
			 * such as function keys or arrow keys, may send special
			 * sequences of characters, such as control-underscore,
			 * plus letters corresponding to modifier keys, plus an
			 * underscore, plus carriage return, which can be used by
			 * the main program for "macro" triggers.  This action
			 * should handle as many events as is efficiently possible
			 * but is only required to handle a single event, and then
			 * only if one is ready or "v" is true.
			 *
			 * This action is required.
			 */

			return (0);
		}

		case TERM_XTRA_FLUSH:
		{
			/*
			 * Flush all pending events XXX XXX XXX
			 *
			 * This action should handle all events waiting on the
			 * queue, optionally discarding all "keypress" events,
			 * since they will be discarded anyway in "z-term.c".
			 *
			 * This action is required, but may not be "essential".
			 */

			return (0);
		}

		case TERM_XTRA_CLEAR:
		{
			int i, j;
			for (i = 0; i < 24; i++) {
				for (j = 0; j < 80; j++) {
					psp_put_char(j * 6, i * 8, 0xFFFFFF, 0x000000, ' ');
				}
			}
			/*
			 * Clear the entire window XXX XXX XXX
			 *
			 * This action should clear the entire window, and redraw
			 * any "borders" or other "graphic" aspects of the window.
			 *
			 * This action is required.
			 */

			return (0);
		}

		case TERM_XTRA_SHAPE:
		{
			/*
			 * Set the cursor visibility XXX XXX XXX
			 *
			 * This action should change the visibility of the cursor,
			 * if possible, to the requested value (0=off, 1=on)
			 *
			 * This action is optional, but can improve both the
			 * efficiency (and attractiveness) of the program.
			 */

			return (0);
		}

		case TERM_XTRA_FROSH:
		{
			/*
			 * Flush a row of output XXX XXX XXX
			 *
			 * This action should make sure that row "v" of the "output"
			 * to the window will actually appear on the window.
			 *
			 * This action is optional, assuming that "Term_text_xxx()"
			 * (and similar functions) draw directly to the screen, or
			 * that the "TERM_XTRA_FRESH" entry below takes care of any
			 * necessary flushing issues.
			 */

			return (0);
		}

		case TERM_XTRA_FRESH:
		{
			/*
			 * Flush output XXX XXX XXX
			 *
			 * This action should make sure that all "output" to the
			 * window will actually appear on the window.
			 *
			 * This action is optional, assuming that "Term_text_xxx()"
			 * (and similar functions) draw directly to the screen, or
			 * that the "TERM_XTRA_FROSH" entry above takes care of any
			 * necessary flushing issues.
			 */

			return (0);
		}

		case TERM_XTRA_NOISE:
		{
			/*
			 * Make a noise XXX XXX XXX
			 *
			 * This action should produce a "beep" noise.
			 *
			 * This action is optional, but convenient.
			 */

			return (0);
		}

		case TERM_XTRA_SOUND:
		{
			/*
			 * Make a sound XXX XXX XXX
			 *
			 * This action should produce sound number "v", where the
			 * "name" of that sound is "sound_names[v]".  This method
			 * is still under construction.
			 *
			 * This action is optional, and not very important.
			 */

			return (0);
		}

		case TERM_XTRA_BORED:
		{
			/*
			 * Handle random events when bored XXX XXX XXX
			 *
			 * This action is optional, and normally not important
			 */

			return (0);
		}

		case TERM_XTRA_REACT:
		{
			/*
			 * React to global changes XXX XXX XXX
			 *
			 * For example, this action can be used to react to
			 * changes in the global "angband_color_table[256][4]" array.
			 *
			 * This action is optional, but can be very useful for
			 * handling "color changes" and the "arg_sound" and/or
			 * "arg_graphics" options.
			 */

			return (0);
		}

		case TERM_XTRA_ALIVE:
		{
			/*
			 * Change the "hard" level XXX XXX XXX
			 *
			 * This action is used if the program changes "aliveness"
			 * by being either "suspended" (v=0) or "resumed" (v=1)
			 * This action is optional, unless the computer uses the
			 * same "physical screen" for multiple programs, in which
			 * case this action should clean up to let other programs
			 * use the screen, or resume from such a cleaned up state.
			 *
			 * This action is currently only used by "main-gcu.c",
			 * on UNIX machines, to allow proper "suspending".
			 */

			return (0);
		}

		case TERM_XTRA_LEVEL:
		{
			/*
			 * Change the "soft" level XXX XXX XXX
			 *
			 * This action is used when the term window changes "activation"
			 * either by becoming "inactive" (v=0) or "active" (v=1)
			 *
			 * This action can be used to do things like activate the proper
			 * font / drawing mode for the newly active term window.  This
			 * action should NOT change which window has the "focus", which
			 * window is "raised", or anything like that.
			 *
			 * This action is optional if all the other things which depend
			 * on what term is active handle activation themself, or if only
			 * one "term_data" structure is supported by this file.
			 */

			return (0);
		}

		case TERM_XTRA_DELAY:
		{
			/*
			 * Delay for some milliseconds XXX XXX XXX
			 *
			 * This action is useful for proper "timing" of certain
			 * visual effects, such as breath attacks.
			 *
			 * This action is optional, but may be required by this file,
			 * especially if special "macro sequences" must be supported.
			 */

			sceKernelDelayThread(v * 1000);
						
			return (0);
		}
	}

	/* Unknown or Unhandled action */
	return (1);
}


/*
 * Display the cursor
 *
 * This routine should display the cursor at the given location
 * (x,y) in some manner.  On some machines this involves actually
 * moving the physical cursor, on others it involves drawing a fake
 * cursor in some form of graphics mode.  Note the "soft_cursor"
 * flag which tells "z-term.c" to treat the "cursor" as a "visual"
 * thing and not as a "hardware" cursor.
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * You may use the "Term_what(x, y, &a, &c)" function, if needed,
 * to determine what attr/char should be "under" the new cursor,
 * for "inverting" purposes or whatever.
 */

bool curs_enable = 0;
bool curs_on = 0;
int curs_x, curs_y;
unsigned char curs_back_attr;
char curs_back_char;

static errr Term_curs_psp(int x, int y)
{
	//term_data *td = (term_data*)(Term->data);
	/* XXX XXX XXX */

	curs_enable = 1;
	curs_x = x;
	curs_y = y;
	curs_on = 1;
	Term_what(x, y, &curs_back_attr, &curs_back_char);
				
	psp_put_char(x * 6, y * 8, 0xFFFFFF, 0x000000, 219); 

	/* Success */
	return (0);
}

void psp_cursor_flip() {
	if (curs_enable) {
		if (curs_on) {
			psp_put_char(curs_x * 6, curs_y * 8, pal[curs_back_attr & 0x0F], pal[curs_back_attr >> 4], curs_back_char);
			curs_on = 0;
		}
		else {
			psp_put_char(curs_x * 6, curs_y * 8, 0xFFFFFF, 0x000000, 219);
			curs_on = 1;
		}
	}
}

/*
 * Erase some characters
 *
 * This function should erase "n" characters starting at (x,y).
 *
 * You may assume "valid" input if the window is properly sized.
 */
static errr Term_wipe_psp(int x, int y, int n)
{
//	term_data *td = (term_data*)(Term->data);

	/* XXX XXX XXX */

	int i;

	for (i = 0; i < n; i++) {
		psp_put_char((x + i) * 6, y * 8, pal[0], 0x000000, 219); // 219 is solid block
	}
	/* Success */
	return (0);
}


/*
 * Draw some text on the screen
 *
 * This function should actually display an array of characters
 * starting at the given location, using the given "attribute",
 * and using the given string of characters, which contains
 * exactly "n" characters and which is NOT null-terminated.
 *
 * You may assume "valid" input if the window is properly sized.
 *
 * You must be sure that the string, when written, erases anything
 * (including any visual cursor) that used to be where the text is
 * drawn.  On many machines this happens automatically, on others,
 * you must first call "Term_wipe_xxx()" to clear the area.
 *
 * In color environments, you should activate the color contained
 * in "color_data[a & 0x0F]", if needed, before drawing anything.
 *
 * You may ignore the "attribute" if you are only supporting a
 * monochrome environment, since this routine is normally never
 * called to display "black" (invisible) text, including the
 * default "spaces", and all other colors should be drawn in
 * the "normal" color in a monochrome environment.
 *
 * Note that if you have changed the "attr_blank" to something
 * which is not black, then this function must be able to draw
 * the resulting "blank" correctly.
 *
 * Note that this function must correctly handle "black" text if
 * the "always_text" flag is set, if this flag is not set, all the
 * "black" text will be handled by the "Term_wipe_xxx()" hook.
 */
static errr Term_text_psp(int x, int y, int n, byte a, const char *cp)
{
//	term_data *td = (term_data*)(Term->data);

	/* XXX XXX XXX */
	
	int i;
	for (i = 0; i < n; i++) {
		if (curs_x == x + i && curs_y == y) curs_enable = 0;
		psp_put_char((x + i) * 6, y * 8, pal[a & 0x0F], 0x000000, cp[i]);
	}
	/* Success */
	return (0);
}


/*
 * Draw some attr/char pairs on the screen
 *
 * This routine should display the given "n" attr/char pairs at
 * the given location (x,y).  This function is only used if one
 * of the flags "always_pict" or "higher_pict" is defined.
 *
 * You must be sure that the attr/char pairs, when displayed, will
 * erase anything (including any visual cursor) that used to be at
 * the given location.  On many machines this is automatic, but on
 * others, you must first call "Term_wipe_psp(x, y, 1)".
 *
 * With the "higher_pict" flag, this function can be used to allow
 * the display of "pseudo-graphic" pictures, for example, by using
 * the attr/char pair as an encoded index into a pixmap of special
 * "pictures".
 *
 * With the "always_pict" flag, this function can be used to force
 * every attr/char pair to be drawn by this function, which can be
 * very useful if this file can optimize its own display calls.
 *
 * This function is often associated with the "arg_graphics" flag.
 *
 * This function is only used if one of the "higher_pict" and/or
 * "always_pict" flags are set.
 */
static errr Term_pict_psp(int x, int y, int n, const byte *ap, const char *cp,
                          const byte *tap, const char *tcp)
{
//	term_data *td = (term_data*)(Term->data);
	/* XXX XXX XXX */

	/* Success */
	return (0);
}



/*** Internal Functions ***/


/*
 * Instantiate a "term_data" structure
 *
 * This is one way to prepare the "term_data" structures and to
 * "link" the various informational pieces together.
 *
 * This function assumes that every window should be 80x24 in size
 * (the standard size) and should be able to queue 256 characters.
 * Technically, only the "main screen window" needs to queue any
 * characters, but this method is simple.  One way to allow some
 * variation is to add fields to the "term_data" structure listing
 * parameters for that window, initialize them in the "init_xxx()"
 * function, and then use them in the code below.
 *
 * Note that "activation" calls the "Term_init_xxx()" hook for
 * the "term" structure, if needed.
 */
static void term_data_link(int i)
{
	term_data *td = &data[i];
	term *t = &td->t;

	/* Initialize the term */
	term_init(t, 80, 24, 256);

	/* Choose "soft" or "hard" cursor XXX XXX XXX */
	/* A "soft" cursor must be explicitly "drawn" by the program */
	/* while a "hard" cursor has some "physical" existance and is */
	/* moved whenever text is drawn on the screen.  See "z-term.c". */
	 t->soft_cursor = TRUE; 

	/* Avoid the "corner" of the window XXX XXX XXX */
	/* t->icky_corner = TRUE; */

	/* Use "Term_pict()" for all attr/char pairs XXX XXX XXX */
	/* See the "Term_pict_xxx()" function above. */
	/* t->always_pict = TRUE; */

	/* Use "Term_pict()" for some attr/char pairs XXX XXX XXX */
	/* See the "Term_pict_xxx()" function above. */
	/* t->higher_pict = TRUE; */

	/* Use "Term_text()" even for "black" text XXX XXX XXX */
	/* See the "Term_text_xxx()" function above. */
	t->always_text = TRUE; 

	/* Ignore the "TERM_XTRA_BORED" action XXX XXX XXX */
	/* This may make things slightly more efficient. */
	t->never_bored = TRUE; 

	/* Ignore the "TERM_XTRA_FROSH" action XXX XXX XXX */
	/* This may make things slightly more efficient. */
	t->never_frosh = TRUE; 

	/* Erase with "white space" XXX XXX XXX */
	t->attr_blank = TERM_DARK; 
	t->char_blank = ' '; 

	/* Prepare the init/nuke hooks */
	t->init_hook = Term_init_psp;
	t->nuke_hook = Term_nuke_psp;

	/* Prepare the template hooks */
	t->user_hook = Term_user_psp;
	t->xtra_hook = Term_xtra_psp;
	t->curs_hook = Term_curs_psp;
	t->wipe_hook = Term_wipe_psp;
	t->text_hook = Term_text_psp;
	t->pict_hook = Term_pict_psp;

	/* Remember where we came from */
	t->data = td;

	/* Activate it */
	Term_activate(t);

	/* Global pointer */
	angband_term[i] = t;
}


#ifdef INTERNAL_MAIN


/*
 * Some special machines need their own "main()" function, which they
 * can provide here, making sure NOT to compile the "main.c" file.
 *
 * These systems usually have some form of "event loop", run forever
 * as the last step of "main()", which handles things like menus and
 * window movement, and calls "play_game(FALSE)" to load a game after
 * initializing "savefile" to a filename, or "play_game(TRUE)" to make
 * a new game.  The event loop would also be triggered by "Term_xtra()"
 * (the TERM_XTRA_EVENT action), in which case the event loop would not
 * actually "loop", but would run once and return.
 */


/*
 * Init some stuff
 *
 * This function is used to keep the "path" variable off the stack.
 */

static void init_stuff(void)
{
	char path[1024];

	/* Use the angband_path, or a default */
	my_strcpy(path, DEFAULT_PATH, sizeof(path));

	/* Make sure it's terminated */
	path[511] = '\0';

	/* Hack -- Add a path separator (only if needed) */
	if (!suffix(path, PATH_SEP)) my_strcat(path, PATH_SEP, sizeof(path));

	/* Initialize */
	init_file_paths(path);
}


void psp_screen_init() {
	vram_base = (void*) (0x40000000 | sceGeEdramGetAddr());
	sceDisplaySetMode(0, PSP_SCREEN_WIDTH, PSP_SCREEN_HEIGHT);
	sceDisplaySetFrameBuf((void*) vram_base, PSP_LINE_SIZE, PSP_PIXEL_FORMAT, 1);
	screen_init = 1;
}

void psp_put_char(int x, int y, unsigned long color, unsigned long bk_color, unsigned char ch) {
	if (!screen_init) return;

	unsigned long* vram = vram_base + y * PSP_LINE_SIZE + x;
	unsigned char* font_ptr = &font[(int)ch * 8];
	
	int i, j;
	for (i = 0; i < 8; i++, font_ptr++) {
		unsigned long* vram_ptr = vram;
		for (j = 0; j < 6; j++) {
			if ((*font_ptr & (128 >> j))) 
				*vram_ptr++ = color;
			else
				*vram_ptr++ = bk_color;
		}
		vram += PSP_LINE_SIZE;
	}
}

void psp_put_string(int x, int y, unsigned long color, unsigned long bk_color, char* string) {
	int i;
	for (i = 0; i < strlen(string); i++) {
		psp_put_char(x + 6 * i, y, color, bk_color, string[i]);
	}
}

int psp_exit_callback(void)
{
	sceKernelTerminateThread(main_thid);
	sceKernelTerminateThread(input_thid);
	sceKernelTerminateThread(timer_thid);
	sceKernelExitGame();
	return 0;
}

void psp_callback_thread(void *arg)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", psp_exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);
	
	sceKernelSleepThreadCB();
}

void psp_timer_thread(void* arg)
{
	while (1) {
		psp_cursor_flip();
		sceKernelDelayThread(166666);
	}
}

int psp_analog_octant(int x, int y) {
	if (x < 83) { // left
		if (y < 83) return STICK_NW;
		if (y < 167) return STICK_W;
		return STICK_SW;
	}
	else if (x < 167) { // center
		if (y < 83) return STICK_N;
		if (y < 167) return STICK_C;
		return STICK_S;
	}
	else { // right
		if (y < 83) return STICK_NE;
		if (y < 167) return STICK_E;
		return STICK_SE;
	}
}

void psp_update_keys() {
	SceCtrlData pad;
	sceCtrlReadBufferPositive(&pad, 1); 

	int i;
	for (i = 0; i < NUM_KEYS; i++) { 
		if (pad.Buttons & key_signals[i]) key_state[i]++; 
		else key_state[i] = -1; 
	} 

	int new_octant = psp_analog_octant(pad.Lx, pad.Ly); 
	if (new_octant == analog_octant) analog_state++; 
	else analog_state = -3; // -3 gives some hysteresis of sorts between octants

	analog_octant = new_octant; 				
}

void psp_show_osk() {
				
	int i, j;
	for (i = 0; i < 4; i++) {
		for (j = 0; j < 13; j++) {
			psp_put_char(201 + 6 * j, (28 + i) * 8, 0xFFFFFF, 0x000000, keys[HELD(KEY_RTRIGGER) ? 1 : 0][j + i * 13]);
		}
	}

	psp_put_string(207, 216, 0xFFFFFF, 0x000000, " Esc ");
	psp_put_string(243, 216, 0xFFFFFF, 0x000000, "Bkspc");
	psp_put_string(207, 256, 0xFFFFFF, 0x000000, "Space");
	psp_put_string(243, 256, 0xFFFFFF, 0x000000, "Enter");
}

void psp_highlight_osk(int cx, int cy, int color) {
	if (!cy) { // top row -> ESC, BKSPC
		if (cx < 6) psp_put_string(207, 216, 0xFFFFFF, color, " Esc ");
		else  psp_put_string(243, 216, 0xFFFFFF, color, "Bkspc");	
	}
	else if (cy < 5) { // normal rows
		psp_put_char(201 + 6 * cx, (27 + cy) * 8, 0xFFFFFF, color, keys[HELD(KEY_RTRIGGER) ? 1 : 0][cx + (cy-1) * 13]);
	}
	else { // bottom row -> SPACE, ENTER
		if (cx < 6) psp_put_string(207, 256, 0xFFFFFF, color, "Space");
		else psp_put_string(243, 256, 0xFFFFFF, color, "Enter");
	}
	pspDebugScreenSetBackColor(0x000000);
}

void psp_hide_osk() {
	int i, j;
	for (i = 0; i < 6; i++) {
		for (j = 0; j < 13; j++) {
			psp_put_char(201 + 6 * j, (27 + i) * 8, 0xFFFFFF, 0x000000, ' ');
		}
	}
}

char psp_do_osk() {
	static int cx = 6; // static so we remember the last cursor position
	static int cy = 3;

	psp_show_osk();
			
	while (HELD(KEY_TRIANGLE)) { // Maintain keyboard until triangle released
		int oldrt = HELD(KEY_RTRIGGER);
		int oldcx = cx;
		int oldcy = cy;
		
		psp_update_keys();		

		/* Highlight the selected whatever */
		psp_highlight_osk(cx, cy, 0xA05000);
		
		
		/* Evaluate keys and respond */

		if (cy > 0 && cy < 5) {
			if (PRESSED(KEY_LEFT)  && --cx < 0)  cx = 12;
			if (PRESSED(KEY_RIGHT) && ++cx > 12) cx = 0;
		}
		else { 
			if (PRESSED(KEY_LEFT) || PRESSED(KEY_RIGHT)) { 
				if (cx < 6) cx += 6;
				else cx -= 6;
			}
		}
		if (PRESSED(KEY_UP)    && --cy < 0)  cy = 5;
		if (PRESSED(KEY_DOWN)  && ++cy > 5)  cy = 0;
		
		if (PRESSED(KEY_CIRCLE)) { // Circle cancels the keyboard
			while (HELD(KEY_TRIANGLE)) psp_update_keys(); // wait until the triangle is released
			return 0; // and then exit with no key
		}
			
		/* If the trigger state changed, draw the new letter set */
		if (HELD(KEY_RTRIGGER) != oldrt) psp_show_osk();
		
		/* If something changed, erase the old highlight */	
		if (cx != oldcx || cy != oldcy || HELD(KEY_RTRIGGER) != oldrt)
			psp_highlight_osk(oldcx, oldcy, 0x000000);

	}
			
	psp_hide_osk();

	if (!cy) {
		if (cx < 6) return 0x1B; // esc
		else return 0x08; // backspace
	}
	else if (cy < 5) {
		char ascii_value = keys[HELD(KEY_RTRIGGER) ? 1 : 0][cx + (cy-1) * 13];

		// Attempt to send ctrl modifier if left trigger held
		if (HELD(KEY_LTRIGGER)) {
			char ctrl_ascii_value = KTRL(ascii_value);
			if (ctrl_ascii_value >= KTRL('A') && ctrl_ascii_value <= KTRL('Z'))
				return KTRL(ascii_value); // but only if a valid input results
			else
				return 0;
		}
		else return ascii_value;
	}
	else {
		if (cx < 6) return ' '; // space
		else return 0x0D; // enter
	}
}

void psp_show_star(star_t* star) {
	if (star->links[0]) {
		psp_put_string(240 - 3 * strlen(star->desc),           232, 0xFFFFFF, 0x000000, star->desc);
		psp_put_string(120 - 3 * strlen(star->links[0]->desc), 232, 0xCC66FF, 0x000000, star->links[0]->desc);
		psp_put_string(240 - 3 * strlen(star->links[1]->desc), 216, 0x66FFFF, 0x000000, star->links[1]->desc);
		psp_put_string(360 - 3 * strlen(star->links[2]->desc), 232, 0x66FF33, 0x000000, star->links[2]->desc);
		psp_put_string(240 - 3 * strlen(star->links[3]->desc), 248, 0xFF6633, 0x000000, star->links[3]->desc);
	}
}

void psp_hide_star() {
	psp_put_string(0, 216, 0x000000, 0x000000, "                                                                                ");
	psp_put_string(0, 232, 0x000000, 0x000000, "                                                                                ");
	psp_put_string(0, 248, 0x000000, 0x000000, "                                                                                ");
}

void psp_save_star(int fff, star_t* star) {
	// Write both strings with their null terminator (as a separator)
	sceIoWrite(fff, &star->macro, strlen(&star->macro[0]) + 1);
	sceIoWrite(fff, &star->desc, strlen(&star->desc[0]) + 1);
				
	if (star->links[0] != NULL) {
		psp_save_star(fff, star->links[0]);
		psp_save_star(fff, star->links[1]);
		psp_save_star(fff, star->links[2]);
		psp_save_star(fff, star->links[3]);
	}
}

void psp_save_macros() {
	int fff = sceIoOpen("ms0:/PSP/GAME/ANGBAND/macros.txt", PSP_O_WRONLY | PSP_O_TRUNC | PSP_O_CREAT, 0777);
	psp_save_star(fff, &star);
	sceIoClose(fff);
}

void psp_load_star(int fff, star_t* star) {

	char c;				
	char* text_ptr = &star->macro[0];
	do {
		sceIoRead(fff, &c, 1);
		*text_ptr++ = c;
	} while (c != 0);	

	text_ptr = &star->desc[0];
	do {
		sceIoRead(fff, &c, 1);
		*text_ptr++ = c;
	} while (c != 0);	
	
	if (star->links[0] != NULL) {
		psp_load_star(fff, star->links[0]);
		psp_load_star(fff, star->links[1]);
		psp_load_star(fff, star->links[2]);
		psp_load_star(fff, star->links[3]);
	}
}

void psp_load_macros() {
	int fff = sceIoOpen("ms0:/PSP/GAME/ANGBAND/macros.txt", PSP_O_RDONLY, 0777);

	if (fff < 0) { // if the file was not there,
		return; // just use the defaults
	}
	
	psp_load_star(fff, &star);
	sceIoClose(fff);				
}

void psp_modify_macro() {
  char cmd1[] = "D-pad: Select macro to replace \x0a: Finish \x07: Cancel";
	char cmd2[] = "\x08: Type macro \x0a: Finish \x07: Cancel";
	char cmd3[] = "\x08: Type description \x0a: Finish \x07: Cancel";
	char cmd4[] = "Saving change...";
	
	star_t* c_star = &star;
				
	psp_put_string(240 - 3 * strlen(cmd1), 200, 0xC020E0, 0x000000, cmd1);	
	while (c_star->links[0] != NULL) {
		psp_show_star(c_star);

		psp_update_keys();

		if (PRESSED(KEY_CROSS)) break;
		else if (PRESSED(KEY_CIRCLE)) {
			psp_put_string(0, 200, 0x000000, 0x000000, "                                                                                ");
			psp_hide_star();
			return;
		}
		else if (PRESSED(KEY_LEFT))  {c_star = c_star->links[0]; psp_hide_star();}
		else if (PRESSED(KEY_UP))    {c_star = c_star->links[1]; psp_hide_star();}
		else if (PRESSED(KEY_RIGHT)) {c_star = c_star->links[2]; psp_hide_star();}
		else if (PRESSED(KEY_DOWN))  {c_star = c_star->links[3]; psp_hide_star();}
	}
	psp_hide_star();

	char macro[64];
	char* macro_ptr = &macro[0];
	
	psp_put_string(240 - 3 * strlen(cmd2), 200, 0xC020E0, 0x000000, cmd2);
	while (!HELD(KEY_CROSS) && macro_ptr < (&macro[0] + 63)) {
		psp_update_keys();

		if (HELD(KEY_TRIANGLE)) {
			char c = psp_do_osk();
			if (c == 0x08 && macro_ptr > &macro[0]) { // backspace
				macro_ptr--;
				*macro_ptr = 0;
				psp_put_string(0, 208, 0x000000, 0x000000, "                                                                                ");
				psp_put_string(240 - 3 * strlen(macro), 208, 0xC020E0, 0x000000, macro);
			}
			else if (c) {
				*macro_ptr++ = c;
				*macro_ptr = 0;
				psp_put_string(240 - 3 * strlen(macro), 208, 0xC020E0, 0x000000, macro);
			}
		}
		else if (HELD(KEY_CIRCLE)) {
			psp_put_string(0, 200, 0x000000, 0x000000, "                                                                                ");
			psp_put_string(0, 208, 0x000000, 0x000000, "                                                                                ");
			return;
		}
	}
	psp_put_string(0, 200, 0x000000, 0x000000, "                                                                                ");
	psp_put_string(0, 208, 0x000000, 0x000000, "                                                                                ");
	
	*macro_ptr = 0;

	// wait for release so one press doesn't count as two
	while (HELD(KEY_CROSS)) psp_update_keys();
	
	char desc[64];
	char* desc_ptr = &desc[0];
	
	psp_put_string(240 - 3 * strlen(cmd3), 200, 0xC020E0, 0x000000, cmd3);
	while (!HELD(KEY_CROSS) && desc_ptr < (&desc[0] + 63)) {
		psp_update_keys();

		if (HELD(KEY_TRIANGLE)) {
			char c = psp_do_osk();
			if (c == 0x08 && desc_ptr > &desc[0]) { // backspace
				desc_ptr--;
				*desc_ptr = 0;
				psp_put_string(0, 208, 0x000000, 0x000000, "                                                                                ");
				psp_put_string(240 - 3 * strlen(desc), 208, 0xC020E0, 0x000000, desc);
			}
			else if (c) {
				*desc_ptr++ = c;
				*desc_ptr = 0;
				psp_put_string(240 - 3 * strlen(desc), 208, 0xC020E0, 0x000000, desc);
			}
		}
		else if (HELD(KEY_CIRCLE)) {
			psp_put_string(0, 200, 0x000000, 0x000000, "                                                                                ");
			psp_put_string(0, 232, 0x000000, 0x000000, "                                                                                ");
			return;
		}
	}
	psp_put_string(0, 200, 0x000000, 0x000000, "                                                                                ");
	psp_put_string(0, 208, 0x000000, 0x000000, "                                                                                ");
	
	*desc_ptr = 0;	

	psp_put_string(240 - 3 * strlen(cmd4), 200, 0xC020E0, 0x000000, cmd4);

	strcpy(c_star->macro, macro);
	strcpy(c_star->desc, desc);
	psp_save_macros();

	psp_put_string(0, 200, 0x000000, 0x000000, "                                                                                ");

}

void psp_do_macro() {
	star_t* c_star = &star;
				
	while (HELD(KEY_SQUARE) && c_star->links[0] != NULL) {
		psp_show_star(c_star);

		psp_update_keys();

		if (PRESSED(KEY_CROSS)) { // change macro
			psp_hide_star();
			psp_modify_macro();
		}
		else if (PRESSED(KEY_CIRCLE)) { // cancel
			while (HELD(KEY_SQUARE)) psp_update_keys();
			psp_hide_star();
			return;
		}
		else if (PRESSED(KEY_LEFT))  {c_star = c_star->links[0]; psp_hide_star();}
		else if (PRESSED(KEY_UP))    {c_star = c_star->links[1]; psp_hide_star();}
		else if (PRESSED(KEY_RIGHT)) {c_star = c_star->links[2]; psp_hide_star();}
		else if (PRESSED(KEY_DOWN))  {c_star = c_star->links[3]; psp_hide_star();}
	}

	psp_hide_star();

	int i;
	for (i = 0; i < strlen(c_star->macro); i++) {
		Term_keypress(c_star->macro[i]);
	}	

	while (HELD(KEY_SQUARE)) psp_update_keys();
}

#define TRY_RUN if (HELD(KEY_RTRIGGER)) Term_keypress('.')

void psp_input_thread(void* arg)
{
				
	int i;
	for (i = 0; i < NUM_KEYS; i++) key_state[i] = -1;

	psp_load_macros();
	
	while (1) {
		psp_update_keys();

		if (HELD(KEY_TRIANGLE)) {
			char c = psp_do_osk();
			if (c) Term_keypress(c);
		}
		else if (HELD(KEY_SQUARE)) psp_do_macro();
		else {
			if (PRESSED(KEY_DOWN) || A_PRESSED(STICK_S))  
				{TRY_RUN; Term_keypress(0x32);} // down
			if (PRESSED(KEY_LEFT) || A_PRESSED(STICK_W)) 
				{TRY_RUN; Term_keypress(0x34);} // left
			if (PRESSED(KEY_UP)   || A_PRESSED(STICK_N)) 
				{TRY_RUN; Term_keypress(0x38);} // up
			if (PRESSED(KEY_RIGHT)|| A_PRESSED(STICK_E)) 
				{TRY_RUN; Term_keypress(0x36);} // right	
			if (A_PRESSED(STICK_NW)) 
				{TRY_RUN; Term_keypress(0x37);}
			if (A_PRESSED(STICK_NE))
				{TRY_RUN; Term_keypress(0x39);}
			if (A_PRESSED(STICK_SW))
				{TRY_RUN; Term_keypress(0x31);}
			if (A_PRESSED(STICK_SE))
				{TRY_RUN; Term_keypress(0x33);}
			if (PRESSED(KEY_CIRCLE)) Term_keypress(0x1B); // esc
			if (PRESSED(KEY_CROSS)) Term_keypress(0x0D); // enter
		}
		
		sceKernelDelayThread(10000); // Sleep for 10ms
	}
}

/*
 * A hook for "quit()".
 *
 * Close down, then fall back into "quit()".
 */
static void quit_hook(cptr s)
{
	int j;

	/* Unused parameter */
	(void)s;

	/* Scan windows */
	for (j = ANGBAND_TERM_MAX - 1; j >= 0; j--)
	{
		/* Unused */
		if (!angband_term[j]) continue;

		/* Nuke it */
		term_nuke(angband_term[j]);
	}
}

int main(int argc, char *argv[])
{

	/* Save the "program name" XXX XXX XXX */
	argv0 = argv[0];

	psp_screen_init();
	pspDebugScreenInit();
	sceCtrlSetSamplingCycle(0);
	sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);

	main_thid = sceKernelCreateThread("update_thread", (SceKernelThreadEntry)psp_callback_thread, 0x11, 0xFA0, 0, 0);
	if (main_thid >= 0) sceKernelStartThread(main_thid, 0, 0);

	input_thid = sceKernelCreateThread("input_thread", (SceKernelThreadEntry)psp_input_thread, 0x11, 0xFA0, 0, 0);
	if (input_thid >= 0) sceKernelStartThread(input_thid, 0, 0);
		
	timer_thid = sceKernelCreateThread("timer_thread", (SceKernelThreadEntry)psp_timer_thread, 0x11, 0xFA0, 0, 0);
	if (timer_thid >= 0) sceKernelStartThread(timer_thid, 0, 0);

	/* Get the file paths */
	init_stuff();

	/* Process the player name */
	process_player_name(TRUE);

	/* Install "quit" hook */
	quit_aux = quit_hook;

	/* Bypass the modules[] stuff from main.c */
	term_data_link(0);
	ANGBAND_SYS = "psp";
				
	/* Initialize */
	init_angband();

	/* Wait for response */
	pause_line(Term->hgt - 1);

	/* Play the game */
	play_game(FALSE);

	/* Free resources */
	cleanup_angband();

	/* Quit */
	sceKernelTerminateThread(main_thid);
	sceKernelTerminateThread(input_thid);
	sceKernelTerminateThread(timer_thid);
	sceKernelExitGame();

	/* Exit */
	return (0);
}

#endif /* INTERNAL_MAIN */

#endif /* USE_PSP */
